/*
Name: H

Version: 1.0
 
Description: computes bibliometric indexes using Google Scholar.

Known limitations: Google Scholar outputs at most 1000 items per query. For queries
that retrieve more than 1000 items the computed indexes may not be fully accutate.

Author: Massimo Franceschet (massimo.franceschet@dimi.uniud.it) - Dipartimento di Matematica e Informatica, Universita' degli Studi di Udine, Via delle Scienze, 206 - 33100 Udine (Italy)

License: GNU General Public License (GPL)    
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

import java.net.*;
import java.io.*;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.util.*;

public class Pagella {
    
  public static void main(String[] args) {    
    
    checkUsage(args);   

    /*
    // offline
    BufferedReader in = null;    
    String google = args[0];
    try {
      in = new BufferedReader(new FileReader(google));
    } catch (IOException ioe) {
      System.exit(1);
    } 
    int items = numberOfItems(in);
    ArrayList<String> citesList = new ArrayList<String>();
    citeFinder(in, citesList);
    */
    
    // make first request (100 items)
    String request = makeRequest(args, 0);
    BufferedReader in = null;    
    try {
      HttpURLConnection con = 
          (HttpURLConnection) (new URL(request)).openConnection();
      
      con.setRequestProperty("User-Agent","H");

      in = new BufferedReader(new InputStreamReader(con.getInputStream()));

    } catch (IOException ioe) {
      System.err.println(ioe);
      System.exit(1);
    }
    

    // find total number of items
    int items = numberOfItems(in);
    
    // retrieve citations from first 100 items
    ArrayList<String> citesList = new ArrayList<String>();
    citeFinder(in, citesList);

    // request next items
    int n = items;
    n = n - 100;
    int start = 100;
    if (items > 100) {
      System.out.print("Please wait: ");
    }  
    while ((start < 1000) && (n > 0)) {
		  System.out.print(".");
		  request = makeRequest(args, start);
		  in = null;    
		  try {
		    HttpURLConnection con = 
		        (HttpURLConnection) (new URL(request)).openConnection();
		    
		    con.setRequestProperty("User-Agent","H");

		    in = new BufferedReader(new InputStreamReader(con.getInputStream()));

		  } catch (IOException ioe) {
		    System.err.println(ioe);
		    System.exit(1);
		  }
	    citeFinder(in, citesList);
	    start = start + 100;
	    n = n - 100;
    }
    if (items > 100) {
      System.out.println(" Done!");
    }  
    
    // compute statistics
    statsMaker(citesList, items);

  
  }
 

  /** find number of items in the input stream */
  public static int numberOfItems(BufferedReader in) {
    
    String regex = 
      "Results\\s*<b>\\d+</b>\\s*-\\s*<b>\\d+</b>(\\s|\\w)*<b>\\d*,?\\d*,?\\d*</b>";
    Pattern pattern = Pattern.compile(regex); 
    Matcher matcher;
    String results, line="";
    int items = 0;
    boolean found = false;
    try {
      while (!found && ((line = in.readLine()) != null)) {
        matcher = pattern.matcher(line);
        while(matcher.find()) {
          results = matcher.group();
          results = results.
            replaceFirst("Results\\s*<b>\\d+</b>\\s*-\\s*<b>\\d+</b>(\\s|\\w)*","");
          results = results.
            replaceAll("<b>","");
          results = results.
            replaceAll("</b>","");
          results = results.
            replaceAll(",","");
          items = Integer.parseInt(results);
          found = true;  
	      }  
      }
    } catch (IOException ioe) {
      System.err.println(ioe);
      System.exit(1);
    }
    
    return items;
  }
      
  /** find number of citations in the input stream */
  public static void citeFinder(BufferedReader in, ArrayList<String> citesList) {
    
    String regex = "Cited\\s*by\\s*\\d+";
    Pattern pattern = Pattern.compile(regex); 
    String citedBy, line = "";
    Matcher matcher;
  
    try {
      while ((line = in.readLine()) != null) {
        // find citation count for each item
        matcher = pattern.matcher(line);
        while(matcher.find()) {
          citedBy = matcher.group();
          citedBy = citedBy.replaceFirst("Cited\\s*by\\s*","");
          citesList.add(citedBy);
        }
      }
    } catch (IOException ioe) {
      System.err.println(ioe);
      System.exit(1);
    }
    
  }
      
  /** compute bibliometric indexes */
  public static void statsMaker(ArrayList<String> citesList, int items) {
    
    
    if (items == 0) {
      System.out.println("Your search did not match any articles.");
      return;
    }
    
    if (items > 1000) {
      System.out.print("WARNING: The number of items selected by the query is " + items + 
      ", but Google Scholar outputs at most 1000 items. ");   
      System.out.println("Computed indexes may be not fully accurate!");   
      System.out.println();
      items = 1000;
    }

    
    int[] cites = new int[citesList.size()];
    for (int i = 0; i < cites.length; i++) {
      cites[i] = Integer.parseInt(citesList.get(i));
    }
    
    insertionSort(cites);
    
    // h index
    int hIndex;
    int h = 1;
    for (int i = 0; i < cites.length; i++) {
      if (cites[i] >= h) h++; else break;
    }
    hIndex = h-1;
    

    // g index
    int gIndex;
    int g = 1, sumCites = 0;
    for (int i = 0; i < cites.length; i++) {
      sumCites = sumCites + cites[i];
      if (sumCites >= (g * g)) g++; else break;
    }
    gIndex = g - 1;
    
    // cited papers and citations 
    int cited = cites.length;
    double perCited = (((double) cited) / items) * 100;
    int cit = 0;
    for (int i = 0; i < cited; i++) {
      cit = cit + cites[i];
    }
    double citesPerItem = ((double) cit) / items;
    double citesPerCitedItem = ((double) cit) / cited;
    
    // print indexes
    System.out.println("items: " + items);   
    System.out.format("cited items: %1$d (%2$.0f%% of items)\n", cited, perCited);
    System.out.println("citations: " + cit);   
    System.out.format("citations per item: %.2f\n", citesPerItem);
    System.out.format("citations per cited item: %.2f\n", citesPerCitedItem);
    System.out.println("h: " + hIndex);
    System.out.print("h core:");
    for (int i = 0; i < hIndex; i++) {
      System.out.print(" " + cites[i]);
    }
    System.out.println();
    System.out.println("g: " + gIndex);
    System.out.print("g core:");
    for (int i = 0; i < gIndex; i++) {
      System.out.print(" " + cites[i]);
    }
    System.out.println();
    
  }



  
  /** build HTTP request */
  public static String makeRequest(String[] args, int startItem) {
      
    String author = "", publication = "", and = "", or = "", not = "", 
           subj = "", start = "", end = "";

    int nargs = args.length;

    try {
      if (args[0].equals("-s")) {
        author = URLEncoder.encode("\"" + args[1] + "\"", "UTF8");
      }  
      if (args[0].equals("-p")) {
        publication = URLEncoder.encode("\"" + args[1] + "\"", "UTF8");
      }  

	    int i = 2;
	    while (i < nargs) {
	      if ((i < nargs) && (args[i].equals("-and"))) {
	        int j = i + 1;
	        while ((j < nargs) && !findOption(args[j])) {
	          if (j == (i + 1)) {
	            and = URLEncoder.encode(args[j], "UTF8");
	          } else {
	            and = and + "+" + URLEncoder.encode(args[j], "UTF8");
	          }
	          j++;
	        }
	        i = j;
	      }
	      
	      if ((i < nargs) && (args[i].equals("-or"))) {
	        int j = i + 1;
	        while ((j < nargs) && !findOption(args[j]))   {
	          if (j == (i + 1)) {
	            or = URLEncoder.encode(args[j], "UTF8");
	          } else {
	            or = or + "+" + URLEncoder.encode(args[j], "UTF8");
	          }
	          j++;
	        }
	        i = j;
	      }
	      
	      if ((i < nargs) && (args[i].equals("-not"))) {
	        int j = i + 1;
	        while ((j < nargs) && !findOption(args[j]))   {
	          if (j == (i + 1)) {
	            not = URLEncoder.encode(args[j], "UTF8");
	          } else {
	            not = not + "+" + URLEncoder.encode(args[j], "UTF8");
	          }
	          j++;
	        }
	        i = j;
	      }
	      
	      if ((i < nargs) && (args[i].equals("-start"))) {
          int j = i + 1;
          if ((j < nargs) && !findOption(args[j])) {
            start = URLEncoder.encode(args[j], "UTF8");
            j++;
          }  
          i = j;
	      }

	      if ((i < nargs) && (args[i].equals("-end"))) {
          int j = i + 1;
          if ((j < nargs) && !findOption(args[j])) {
            end = URLEncoder.encode(args[j], "UTF8");
            j++;
          }  
          i = j;
	      }

	      if ((i < nargs) && (args[i].equals("-subj"))) {
          int j = i + 1;
          if ((j < nargs) && !findOption(args[j])) {
            subj = URLEncoder.encode(args[j], "UTF8");
            j++;
          }  
          i = j;
	      }

	      if ((i < nargs) && !findOption(args[i]))   {
	        int j = i;
	        while (j < nargs) {
	          if (j == i) {
	            and = URLEncoder.encode(args[j], "UTF8");
	          } else {
	            and = and + "+" + URLEncoder.encode(args[j], "UTF8");
	          }
	          j++;
	        }
	        i = j;
	      }	        
	    }
    
    } catch (Exception e) {
      System.err.println(e);
      System.exit(1);
    }
 
    String s = "";
    if (subj.equals("")) {
      s = "&as_allsubj=all";
    } else {
      s = "&as_allsubj=some&as_subj=" + subj;
    }  
    
    String request = 
        "http://scholar.google.com/scholar?" + 
        "as_q=" + and + 
        "&num=100" + 
        "&btnG=Search+Scholar" + 
        "&as_epq=" + 
        "&as_oq=" + or +
        "&as_eq=" + not + 
        "&as_occt=any" + 
        "&as_sauthors=" + author + 
        "&as_publication=" + publication + 
        "&as_ylo=" + start +
        "&as_yhi=" + end +
        s + 
        "&hl=en" +
        "&lr=" + 
        "&safe=off";
    
    if (startItem > 0) { 
      request = request + "&start=" + startItem + "&sa=N";
	  }      

    return request;  
  
  }
  
  
  /** find an option in command  */
  public static boolean findOption(String opt) {
   
    return ((opt.equals("-and"))   || 
		        (opt.equals("-or"))    ||
		        (opt.equals("-not"))   || 
		        (opt.equals("-start")) || 
		        (opt.equals("-end")) || 
		        (opt.equals("-subj")) 
		        );   
  
  }



  
  /** sort in decreasing order */
  public static void insertionSort(int[] a) {
  
    int i, j, key;
    
    int len = a.length;
    for (j = 1; j < len; j++) {
      key = a[j];
      i = j-1;
      while ((i >= 0) && (key > a[i])) {
        a[i+1] = a[i];
        i--;
      }
      a[i+1] = key;      
    }      
  }
  
  /** print program's usage */
  public static void printUsage() {
    
    System.out.println("Usage: java Pagella (-s scholar | -p publication) [options]");
    System.out.println();
		System.out.println("Use one of the two following MODES:"); 
		System.out.println("-s scholar: to search for a scholar");
		System.out.println("-p publication: to search for a publication source");
    System.out.println();
		System.out.println("You may add an option to refine your search. Available OPTIONS are:");
		System.out.println("-and words: with all of the words");
		System.out.println("-or words:  with any of the words");
		System.out.println("-not words: with none of the words");
		System.out.println("-start: starting at year start");
		System.out.println("-end: ending at year end");
		System.out.println("-subj: in subject subj. Subjects are:");
		System.out.println("\tbio: Biology, Life Sciences, and Environmental Science");
		System.out.println("\tbus: Business, Administration, Finance, and Economics");
		System.out.println("\tchm: Chemistry and Materials Science");
		System.out.println("\teng: Engineering, Computer Science, and Mathematics");
		System.out.println("\tmed: Medicine, Pharmacology, and Veterinary Science");
		System.out.println("\tphy: Physics, Astronomy, and Planetary Science");
		System.out.println("\tsoc: Social Sciences, Arts, and Humanities");
		
    System.exit(1);
    
  }
  

  /** check command-line input */  
  public static void checkUsage(String[] args) {
    
    int nargs = args.length;
    
    if ((nargs < 2) || ((!args[0].equals("-s")) && (!args[0].equals("-p")))) {
      printUsage();
    }
    
    
  }

} // end of class Scholar


